#####################################################################
# Copyright (c) 2008, Impulse Accelerated Technologies, Inc.
# All Rights Reserved.
#
# embeddedTclFile.tcl - Utilities for processing template files with 
#								embedded Tcl for auto generation of files
#
#####################################################################

#####################################################################
# Embedded Tcl file is a text file that also contains pieces of Tcl for
#  generating sections in-line.  Sections begin with a line starting with
#  "&embtcl_begin;" and end with a line starting with "&embtcl_end;".  All 
#	other lines in the file will be outputed as-is.
#  Example:
#		&embtcl_begin;
#			puts $EmbeddedTclVars(OutputFile) "	Name=\"$EmbeddedTclVars(AppName)\""
#		&embtcl_end;

# For automatic generation when copying an entire directory using 
#	copyDirAndGenerateEmbeddedTclFiles(), the Embedded Tcl files in the 
#	directory structure need to end in ".embTcl" and may contain "{<var_name>}"
#  where <var_name> must be defined in global array EmbeddedTclVars.  Output
#	file name will be the anme of the original .embTcl file with ".embTcl" 
#	removed and any substitutions.
# 	Example:
#		set EmbeddedTclVars(AppName) "MyApp"
#		copyDirAndGenerateEmbeddedTclFiles $srcdir $destdir
# 	 A file named "{AppName}.vcproj.embTcl" will be written as "MyApp.vcproj"

#####################################################################
# global EmbeddedTclVars is used to convery data to Tcl script within template file, 
#  and needs to be set up before calling generateEmbeddedTclFile() 
#  Example call sequence:
#		global EmbeddedTclVars						- need to declare as global to correctly reference variable 
#		array unset EmbeddedTclVars 				- empty array to avoid cross-contamination between calls
#		set EmbeddedTclVars(AppName) "MyApp"	- set variable 'AppName' for use by embedded Tcl
#		generateEmbeddedTclFile $srcfile [file join $destdir [resolveEmbeddedVariableFilename $srcfile]] - generate destfile from srcfile
global EmbeddedTclVars
array set EmbeddedTclVars { }
global EmbeddedTclFileDebugLevel
set EmbeddedTclFileDebugLevel 0

proc generateEmbeddedTclFile {srcfile destfile} {
	global EmbeddedTclVars
	global EmbeddedTclFileDebugLevel
	if { $EmbeddedTclFileDebugLevel > 0 } { 
		puts "DBG:generateEmbeddedTclFile():Begin"
		puts "DBG:generateEmbeddedTclFile():srcfile = $srcfile"
		puts "DBG:generateEmbeddedTclFile():destfile = $destfile"
	
		foreach {key val} [array get EmbeddedTclVars] {
			puts "DBG:generateEmbeddedTclFile():Array($key) = $val"
		}
	}
	
	if { [catch {set embTclFileSrc [open $srcfile r]}] } {
		puts "embeddedTclFile.tcl:generateEmbeddedTclFile():ERROR:Cannot open $srcfile for reading" 
		exit -1
	}
	if { [catch {set embTclFileDest [open $destfile w]}] } {
		puts "embeddedTclFile.tcl:generateEmbeddedTclFile():ERROR:Cannot open $destfile for writing" 
		exit -1
	}
	set EmbeddedTclVars(OutputFile) $embTclFileDest
	
	set foundBegin 0
	while { [gets $embTclFileSrc line] >= 0 } {
		#looking for start of embedded Tcl block
		if { $foundBegin == 0 } { 
			if { [string first "\&embtcl_begin\;" $line] == 0 } {
				if { $EmbeddedTclFileDebugLevel > 0 } {puts "DBG:generateEmbeddedTclFile():Found Block beginning"}
				set foundBegin 1
				set embCode ""
			} else {
				# not part of an embedded script, just write out as-is
				puts $EmbeddedTclVars(OutputFile) $line
			}
		} else {
			# found begin, now looking for end
			if { [string first "\&embtcl_end\;" $line] == 0 } {
				if { $EmbeddedTclFileDebugLevel > 0 } {
					puts "DBG:generateEmbeddedTclFile():Found Block ending"
					puts "DBG:generateEmbeddedTclFile():Evaluating block:$embCode"
				}
				if { [catch {eval $embCode}] } {
					puts "--------------------------------------------------------------------------------"
					puts "embeddedTclFile.tcl:generateEmbeddedTclFile():ERROR:Evaluating code in file $srcfile"
					global errorInfo
					puts "errorInfo=$errorInfo"
					puts "--------------------------------------------------------------------------------"
					puts "   Error is in code block:"
					puts "$embCode"
					puts "--------------------------------------------------------------------------------"
					exit -1
				}
				set foundBegin 0
			} else {
				# part of an embedded script, save for later evaluation, prepend ';' and
				#	newline to separate multiple lines and handle comments correctly if 
				#  line continuation '\' is not present, trim and continue line if it is
				if { [string index $line end] == "\\" } {
					append embCode [string range "$line" 0 end-1]
				} else {
					append embCode "$line;\n"
				}
			}
		}
	}
	if { $foundBegin == 1 } {
		puts "embeddedTclFile.tcl:generateEmbeddedTclFile():ERROR:Didn't find 'embtcl_end' in file $srcfile, found:\[$embCode\]"
		exit -1
	}
	close $embTclFileDest
	close $embTclFileSrc
	if { $EmbeddedTclFileDebugLevel > 0 } {puts "DBG:generateEmbTclFile():End"}
}

proc resolveEmbeddedVariableFilename {srcfile} {
	global EmbeddedTclVars
	global EmbeddedTclFileDebugLevel
	# do substitutions if present
	set fnameOut [file rootname $srcfile]
	while (1) {
		set result [regexp {\{(.*?)\}} $fnameOut match varname ]
		if { $result } {
			#puts "copyDirAndGenerateEmbeddedTclFiles():varname = $varname"
			set fnameOut [string map [list \{$varname\} $EmbeddedTclVars($varname)] $fnameOut]
		} else {
			break;
		}
	}
	return $fnameOut
}
	
proc copyDirAndGenerateEmbeddedTclFiles {srcdir destdir} {
	global env
	global EmbeddedTclVars
	global EmbeddedTclFileDebugLevel
	
	if { $EmbeddedTclFileDebugLevel > 0 } {
		puts "DBG:copyDirAndGenerateEmbeddedTclFiles():Begin"
		puts "DBG:copyDirAndGenerateEmbeddedTclFiles():srcdir = $srcdir"
		puts "DBG:copyDirAndGenerateEmbeddedTclFiles():destdir = $destdir"
	}
	
	# get list of files in srcdir, copy files, generate .embTcls' and 
	#  recursively traverse directories
	foreach f [glob -nocomplain -directory $srcdir * ] {
		set fname [file tail $f]
		if { [file isdirectory $f] } {
			if { $EmbeddedTclFileDebugLevel > 0 } {puts "DBG:copyDirAndGenerateEmbeddedTclFiles():SrcFile = $f is dir"}
			if { $fname == "CVS" } {
				# ignore CVS directories
				puts "INFO:copyDirAndGenerateEmbeddedTclFiles():Skipping CVS directory $f"
			} elseif { [string index $f end] == "~" } {
				# ignore backup files from JEdit
				puts "INFO:copyDirAndGenerateEmbeddedTclFiles():Skipping directory $f because it ends in '~'"
			} else {
				# just a directory, just make it resolving name if necessary and recurse into it
				set fnameDir [resolveEmbeddedVariableFilename $fname]
				file mkdir [file join $destdir $fnameDir ]
				copyDirAndGenerateEmbeddedTclFiles [file join $srcdir $fname] [file join $destdir $fnameDir]
			}
		} else {
			if { [string index $f end] == "~" } {
				# ignore backup files from JEdit
				puts "INFO:copyDirAndGenerateEmbeddedTclFiles():Skipping file $f because it ends in '~'"
			} elseif { [file extension $fname] == ".embTcl" } {
				# process embedded Tcl file
				if { $EmbeddedTclFileDebugLevel > 0 } {puts "DBG:copyDirAndGenerateEmbeddedTclFiles():SrcFile = $fname is .embTcl"}
				set fnameOut [resolveEmbeddedVariableFilename $fname]
				if { $EmbeddedTclFileDebugLevel > 0 } {puts "DBG:copyDirAndGenerateEmbeddedTclFiles():DestFile = $fnameOut will be output file"}
				generateEmbeddedTclFile [file join $srcdir $fname] [file join $destdir $fnameOut]
			} else {
				# normal file, just copy it
				if { $EmbeddedTclFileDebugLevel > 0 } {puts "DBG:copyDirAndGenerateEmbeddedTclFiles():SrcFile = $fname"}
				file copy -force [file join $srcdir $fname] [file join $destdir $fname]
			}
		}
	}
	if { $EmbeddedTclFileDebugLevel > 0 } {puts "DBG:copyDirAndGenerateEmbeddedTclFiles():End"}
}
